from typing import get_args


# Necessary so that we can reference Field in FieldMeta
Field = None


class FieldMeta(type):
    def __getitem__(self, item: type):
        if isinstance(item, tuple):
            assert len(item) == 1, "Only 1 item can be passed in the Field typing."
            item = item[0]

        args = get_args(item)
        nullable = None in args
        assert (
            len(args) <= 2 or len(args) == 2 and not nullable
        ), "Only union between a type and None is accepted."

        if args:
            type_ = [arg for arg in args if arg is not None][0]
        else:
            type_ = item

        return Field(type_, nullable=nullable)


class Field(metaclass=FieldMeta):
    def __init__(self, type_: type, nullable: bool = False):
        from .base import Model # TODO: Find a way to avoid this import

        self.type: type = type_
        self.nullable: bool = nullable

        if issubclass(type_, Model):
            self._field = ForeignKeyField
        elif type_ not in FIELD_MAPPING:
            raise TypeError(f"{type_} is not supported.")
        else:
            self._field = FIELD_MAPPING[self.type]

    def __call__(self, **kwargs):
        return self._field(
            python_type=self.type, nullable=self.nullable, **kwargs
        )


class BaseField:
    def __init__(self, python_type=int, nullable=False, **kwargs):
        self.python_type = python_type
        self.nullable = nullable
        self.is_pk = kwargs.get("primary_key", False)


class StringField(BaseField):
    def __init__(self, python_type=str, nullable=False, **kwargs):
        super().__init__(python_type=python_type, nullable=nullable)
        self.max_length = kwargs.get("max_char")


class IntegerField(BaseField):
    pass


class ForeignKeyField(BaseField):
    pass


FIELD_MAPPING = {str: StringField, int: IntegerField}



from typing import Any
from .fields import BaseField, Field, IntegerField


class ModelMeta(type):
    def __new__(
        cls: type,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwds: Any,
    ):
        instance = super().__new__(cls, name, bases, namespace, **kwds)

        if not (__annotations__ := namespace.get("__annotations__")):
            return instance

        pk_fields = []
        for name, obj in __annotations__.items():
            if obj is Field:
                raise TypeError(
                    "Field must be used with type hinting. For example, call it like so: Field[str]."
                )
            if isinstance(obj, Field):
                obj = obj()
            if isinstance(obj, BaseField) and name in namespace:
                obj.default = namespace[name]
            setattr(instance, name, obj)

            if obj.is_pk:
                pk_fields.append(obj)

        if not pk_fields:
            instance.id = IntegerField(primary_key=True)

        return instance


class Model(metaclass=ModelMeta):
    @property
    def fields(self):
        return {
            field_name: field
            for field_name, field in self.__class__.__dict__.items()
            if isinstance(field, BaseField)
        }



from abc import ABCMeta, abstractmethod


class BaseDatabaseClient(metaclass=ABCMeta):
    _instances = None
    _db_name = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance

    def __init__(self, db_name):
        self._db_name = db_name

        self._connection = None

    @abstractmethod
    def get_connection(self):
        pass

    @abstractmethod
    def execute(self, statement, *args, **kwargs):
        pass

    @abstractmethod
    def close(self):
        pass


class BaseDatabaseSchemaEditor:
    pass
